# -*- mode: python; coding: utf-8 -*-

__author__       = 'Simon Portegies Zwart'
__author_email__ = '<spz@science.uva.nl>'
__date__         = '2006-11-11 15:32:00 CEST'

# Dummy file to indicate that this directory is a package.
